@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -24,78 +24,120 @@
 #endif
 
 #include <errno.h>
-#include <unistd.h>
-#include <string.h>
-#include <sys/ioctl.h>
-#include <sys/socket.h>
-#include <linux/if.h>
-#include <linux/netlink.h>
-#include <linux/rtnetlink.h>
+#include <net/if.h>
+
+#ifndef IFF_LOWER_UP
+#define IFF_LOWER_UP	0x10000
+#endif
 
 #include <glib.h>
 
 #define CONNMAN_API_SUBJECT_TO_CHANGE
 #include <connman/plugin.h>
 #include <connman/device.h>
+#include <connman/inet.h>
 #include <connman/rtnl.h>
 #include <connman/log.h>
 
-#include "inet.h"
-
 struct ethernet_data {
 	int index;
 	unsigned flags;
+	unsigned int watch;
+	struct connman_network *network;
 };
 
-static GSList *ethernet_list = NULL;
-
-static void ethernet_newlink(unsigned short type, int index,
-					unsigned flags, unsigned change)
+static int cable_probe(struct connman_network *network)
 {
-	GSList *list;
+	DBG("network %p", network);
 
-	DBG("index %d flags %ld change %ld", index, flags, change);
-
-	for (list = ethernet_list; list; list = list->next) {
-		struct connman_device *device = list->data;
-		struct ethernet_data *ethernet;
-
-		ethernet = connman_device_get_data(device);
-		if (ethernet == NULL)
-			continue;
-
-		if (ethernet->index != index)
-			continue;
-
-		if ((ethernet->flags & IFF_UP) != (flags & IFF_UP)) {
-			if (flags & IFF_UP) {
-				DBG("power on");
-				connman_device_set_powered(device, TRUE);
-			} else {
-				DBG("power off");
-				connman_device_set_powered(device, FALSE);
-			}
-		}
-
-		if ((ethernet->flags & IFF_LOWER_UP) != (flags & IFF_LOWER_UP)) {
-			if (flags & IFF_LOWER_UP) {
-				DBG("carrier on");
-				connman_device_set_carrier(device, TRUE);
-			} else {
-				DBG("carrier off");
-				connman_device_set_carrier(device, FALSE);
-			}
-		}
-
-		ethernet->flags = flags;
-	}
+	return 0;
 }
 
-static struct connman_rtnl ethernet_rtnl = {
-	.name		= "ethernet",
-	.newlink	= ethernet_newlink,
+static void cable_remove(struct connman_network *network)
+{
+	DBG("network %p", network);
+}
+
+static int cable_connect(struct connman_network *network)
+{
+	DBG("network %p", network);
+
+	return 0;
+}
+
+static int cable_disconnect(struct connman_network *network)
+{
+	DBG("network %p", network);
+
+	return 0;
+}
+
+static struct connman_network_driver cable_driver = {
+	.name		= "cable",
+	.type		= CONNMAN_NETWORK_TYPE_ETHERNET,
+	.probe		= cable_probe,
+	.remove		= cable_remove,
+	.connect	= cable_connect,
+	.disconnect	= cable_disconnect,
 };
 
+static void add_network(struct connman_device *device)
+{
+	struct connman_network *network;
+	int index;
+
+	network = connman_network_create("carrier",
+					CONNMAN_NETWORK_TYPE_ETHERNET);
+	if (network == NULL)
+		return;
+
+	index = connman_device_get_index(device);
+	connman_network_set_index(network, index);
+
+	connman_network_set_name(network, "Wired");
+
+	if (connman_device_add_network(device, network) < 0) {
+		connman_network_unref(network);
+		return;
+	}
+
+	connman_network_set_available(network, TRUE);
+
+	connman_network_set_group(network, "cable");
+
+	connman_network_set_connected(network, TRUE);
+}
+
+static void ethernet_newlink(unsigned flags, unsigned change, void *user_data)
+{
+	struct connman_device *device = user_data;
+	struct ethernet_data *ethernet = connman_device_get_data(device);
+
+	DBG("index %d flags %d change %d", ethernet->index, flags, change);
+
+	if ((ethernet->flags & IFF_UP) != (flags & IFF_UP)) {
+		if (flags & IFF_UP) {
+			DBG("power on");
+			connman_device_set_powered(device, TRUE);
+		} else {
+			DBG("power off");
+			connman_device_set_powered(device, FALSE);
+		}
+	}
+
+	if ((ethernet->flags & IFF_LOWER_UP) != (flags & IFF_LOWER_UP)) {
+		if (flags & IFF_LOWER_UP) {
+			DBG("carrier on");
+			add_network(device);
+		} else {
+			DBG("carrier off");
+			connman_device_remove_all_networks(device);
+		}
+	}
+
+	ethernet->flags = flags;
+}
+
 static int ethernet_probe(struct connman_device *device)
 {
 	struct ethernet_data *ethernet;
@@ -106,13 +148,13 @@ static int ethernet_probe(struct connman_device *device)
 	if (ethernet == NULL)
 		return -ENOMEM;
 
-	ethernet_list = g_slist_append(ethernet_list, device);
-
 	connman_device_set_data(device, ethernet);
 
 	ethernet->index = connman_device_get_index(device);
+	ethernet->flags = 0;
 
-	connman_rtnl_send_getlink();
+	ethernet->watch = connman_rtnl_add_newlink_watch(ethernet->index,
+						ethernet_newlink, device);
 
 	return 0;
 }
@@ -125,7 +167,9 @@ static void ethernet_remove(struct connman_device *device)
 
 	connman_device_set_data(device, NULL);
 
-	ethernet_list = g_slist_remove(ethernet_list, device);
+	connman_rtnl_remove_watch(ethernet->watch);
+
+	connman_device_remove_all_networks(device);
 
 	g_free(ethernet);
 }
@@ -136,7 +180,7 @@ static int ethernet_enable(struct connman_device *device)
 
 	DBG("device %p", device);
 
-	return inet_ifup(ethernet->index);
+	return connman_inet_ifup(ethernet->index);
 }
 
 static int ethernet_disable(struct connman_device *device)
@@ -145,7 +189,7 @@ static int ethernet_disable(struct connman_device *device)
 
 	DBG("device %p", device);
 
-	return inet_ifdown(ethernet->index);
+	return connman_inet_ifdown(ethernet->index);
 }
 
 static struct connman_device_driver ethernet_driver = {
@@ -161,13 +205,13 @@ static int ethernet_init(void)
 {
 	int err;
 
-	err = connman_rtnl_register(&ethernet_rtnl);
+	err = connman_network_driver_register(&cable_driver);
 	if (err < 0)
 		return err;
 
 	err = connman_device_driver_register(&ethernet_driver);
 	if (err < 0) {
-		connman_rtnl_unregister(&ethernet_rtnl);
+		connman_network_driver_unregister(&cable_driver);
 		return err;
 	}
 
@@ -176,10 +220,10 @@ static int ethernet_init(void)
 
 static void ethernet_exit(void)
 {
-	connman_device_driver_unregister(&ethernet_driver);
+	connman_network_driver_unregister(&cable_driver);
 
-	connman_rtnl_unregister(&ethernet_rtnl);
+	connman_device_driver_unregister(&ethernet_driver);
 }
 
 CONNMAN_PLUGIN_DEFINE(ethernet, "Ethernet interface plugin", VERSION,
-						ethernet_init, ethernet_exit)
+		CONNMAN_PLUGIN_PRIORITY_DEFAULT, ethernet_init, ethernet_exit)
