@@ -444,11 +444,10 @@ done:
 
 struct connman_device *connman_inet_create_device(int index)
 {
-	enum connman_device_mode mode = CONNMAN_DEVICE_MODE_UNKNOWN;
 	enum connman_device_type type;
 	struct connman_device *device;
 	char *devname, *ident = NULL;
-	char *addr = NULL, *name = NULL, *node = NULL;
+	char *addr = NULL, *name = NULL;
 
 	if (index < 0)
 		return NULL;
@@ -491,30 +490,23 @@ struct connman_device *connman_inet_create_device(int index)
 	case CONNMAN_DEVICE_TYPE_UNKNOWN:
 	case CONNMAN_DEVICE_TYPE_VENDOR:
 	case CONNMAN_DEVICE_TYPE_GPS:
-		mode = CONNMAN_DEVICE_MODE_UNKNOWN;
 		break;
 	case CONNMAN_DEVICE_TYPE_ETHERNET:
-		mode = CONNMAN_DEVICE_MODE_NETWORK_SINGLE;
 		ident = index2ident(index, NULL);
 		break;
 	case CONNMAN_DEVICE_TYPE_WIFI:
 	case CONNMAN_DEVICE_TYPE_WIMAX:
-		mode = CONNMAN_DEVICE_MODE_NETWORK_SINGLE;
 		ident = index2ident(index, NULL);
 		break;
 	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
-		mode = CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE;
 		break;
 	case CONNMAN_DEVICE_TYPE_CELLULAR:
-		mode = CONNMAN_DEVICE_MODE_NETWORK_SINGLE;
 		ident = index2ident(index, NULL);
 		break;
 	}
 
-	connman_device_set_mode(device, mode);
-
 	connman_device_set_index(device, index);
-	connman_device_set_interface(device, devname, node);
+	connman_device_set_interface(device, devname);
 
 	if (ident != NULL) {
 		connman_device_set_ident(device, ident);
@@ -525,7 +517,6 @@ struct connman_device *connman_inet_create_device(int index)
 
 done:
 	g_free(devname);
-	g_free(node);
 	free(name);
 	free(addr);
 
